@@ -590,7 +590,7 @@ execcase(Estate state, int do_exec)
 	    }
 	    if (!(pprog = patcompile(pat, (save ? PAT_ZDUP : PAT_STATIC),
 				     NULL)))
-		zerr("bad pattern: %s", pat, 0);
+		zerr("bad pattern: %s", pat);
 	    else if (save)
 		*spprog = pprog;
 	}
@@ -606,9 +606,10 @@ execcase(Estate state, int do_exec)
 		execlist(state, 1, ((WC_CASE_TYPE(code) == WC_CASE_OR) &&
 				    do_exec));
 	    }
-	    break;
-	} else
-	    state->pc = next;
+	    if (WC_CASE_TYPE(code) != WC_CASE_TESTAND)
+		break;
+	}
+	state->pc = next;
     }
     cmdpop();
 
